@@ -1,22 +1,21 @@
 use core::fmt;
 use std::{
-    fs::File,
-    io::{BufReader, Cursor, Read, Seek, SeekFrom},
+    collections::HashMap,
+    io::{BufRead, Cursor, Seek, SeekFrom},
     sync::{Arc, Mutex, MutexGuard},
 };
 
 use byteorder::{LittleEndian, ReadBytesExt};
-use image::io::Reader as ImageReader;
-use image::{DynamicImage, ImageFormat, RgbaImage};
+use image::ImageFormat;
 use nalgebra::Vector2;
 
 use crate::{
     channel::{AcquisitionChannel, ChannelIdentifier},
+    convert::DCMLocation,
     error::MCDError,
-    images::read_image_data,
     mcd::AcquisitionXML,
     transform::AffineTransform,
-    BoundingBox, ChannelImage, OnSlide, Print,
+    BoundingBox, ChannelImage, OnSlide, OpticalImage, Print, Region,
 };
 
 #[derive(Debug)]
@@ -24,18 +23,39 @@ pub enum DataFormat {
     Float,
 }
 
-#[derive(Debug)]
-pub(crate) struct DataLocation {
-    pub(crate) reader: Arc<Mutex<BufReader<File>>>,
+// #[derive(Debug)]
+// pub struct DataLocation {
+//     pub reader: Arc<Mutex<BufReader<File>>>,
 
-    pub(crate) offsets: Vec<u64>,
-    pub(crate) sizes: Vec<u64>,
+//     pub offsets: Vec<u64>,
+//     pub sizes: Vec<u64>,
+// }
+
+/// AcquisitionIdentifier is a way of identifying a specific acquisition
+#[derive(Debug)]
+pub enum AcquisitionIdentifier {
+    /// Identified by unique identifier
+    Id(u16),
+    /// Identified by the number specifying the order in which the acquisition was acquired
+    Order(i16),
+    /// Match the description of the acquistion (specified by the user)
+    Description(String),
 }
 
-pub enum AcquisitionIdentifier {
-    Id(u16),
-    Order(i16),
-    Description(String),
+impl fmt::Display for AcquisitionIdentifier {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            AcquisitionIdentifier::Id(id) => {
+                write!(f, "acquisition id: {}", id)
+            }
+            AcquisitionIdentifier::Order(order) => {
+                write!(f, "acquisition order: {}", order)
+            }
+            AcquisitionIdentifier::Description(description) => {
+                write!(f, "acquisition description: {}", description)
+            }
+        }
+    }
 }
 
 #[derive(Debug)]
@@ -43,11 +63,41 @@ pub enum ProfilingType {
     Global,
 }
 
+/// Trait describing a collection of acquisitions and providing methods to summarise
+/// the collection (e.g. to get the list of channels)
+pub trait Acquisitions {
+    /// Returns a list of unique channels from the collection of acquisitions
+    fn channels(&self) -> Vec<&AcquisitionChannel>;
+}
+
+impl<T: BufRead + Seek> Acquisitions for Vec<&Acquisition<T>> {
+    fn channels(&self) -> Vec<&AcquisitionChannel> {
+        let mut channels = HashMap::new();
+
+        for acquisition in self {
+            for channel in acquisition.channels() {
+                if !channels.contains_key(channel.name()) {
+                    channels.insert(channel.name(), channel);
+                }
+            }
+        }
+
+        let mut ordered_channels = Vec::new();
+        for (_, channel) in channels.drain() {
+            ordered_channels.push(channel);
+        }
+
+        ordered_channels.sort_by_key(|a| a.label().to_ascii_lowercase());
+
+        ordered_channels
+    }
+}
+
 /// Acquisition represents a single region analysed by IMC.
 #[derive(Debug)]
-pub struct Acquisition<T: Read + Seek> {
+pub struct Acquisition<T: BufRead + Seek> {
     pub(crate) reader: Option<Arc<Mutex<T>>>,
-    pub(crate) dcm_location: Option<DataLocation>,
+    pub(crate) dcm_location: Option<DCMLocation>,
 
     id: u16,
     description: String,
@@ -85,13 +135,13 @@ pub struct Acquisition<T: Read + Seek> {
     channels: Vec<AcquisitionChannel>,
 }
 
-pub struct SpectrumIterator<'a, T: Read + Seek> {
+pub struct SpectrumIterator<'a, T: BufRead + Seek> {
     acquisition: &'a Acquisition<T>,
     reader: MutexGuard<'a, T>,
     buffer: Vec<u8>,
 }
 
-impl<'a, T: Read + Seek> SpectrumIterator<'a, T> {
+impl<'a, T: BufRead + Seek> SpectrumIterator<'a, T> {
     fn new(acquisition: &'a Acquisition<T>) -> Self {
         let mut reader = acquisition.reader.as_ref().unwrap().lock().unwrap();
 
@@ -108,7 +158,7 @@ impl<'a, T: Read + Seek> SpectrumIterator<'a, T> {
     }
 }
 
-impl<'a, T: Read + Seek> Iterator for SpectrumIterator<'a, T> {
+impl<'a, T: BufRead + Seek> Iterator for SpectrumIterator<'a, T> {
     type Item = Vec<f32>;
 
     fn next(&mut self) -> Option<Vec<f32>> {
@@ -132,7 +182,7 @@ impl<'a, T: Read + Seek> Iterator for SpectrumIterator<'a, T> {
     }
 }
 
-impl<T: Read + Seek> Acquisition<T> {
+impl<T: BufRead + Seek> Acquisition<T> {
     /// Returns the ID associated with the acquisition
     pub fn id(&self) -> u16 {
         self.id
@@ -173,7 +223,7 @@ impl<T: Read + Seek> Acquisition<T> {
         self.profiling_type.as_ref()
     }
 
-    fn image_data(&self, start: i64, end: i64) -> Result<Vec<u8>, std::io::Error> {
+    /*fn image_data(&self, start: i64, end: i64) -> Result<Vec<u8>, std::io::Error> {
         let mutex = self
             .reader
             .as_ref()
@@ -187,27 +237,32 @@ impl<T: Read + Seek> Acquisition<T> {
         let mut reader = ImageReader::new(Cursor::new(self.image_data(start, end).unwrap()));
         reader.set_format(ImageFormat::Png);
         reader.decode().unwrap()
-    }
+    }*/
 
     /// Returns the optical image of the acquisition region prior to ablation
-    pub fn before_ablation_image(&self) -> RgbaImage {
-        match self.dynamic_image(
-            self.before_ablation_image_start_offset,
-            self.before_ablation_image_end_offset,
-        ) {
-            DynamicImage::ImageRgba8(rgba8) => rgba8,
-            _ => panic!("Unexpected DynamicImage type"),
+    pub fn before_ablation_image(&self) -> OpticalImage<T> {
+        OpticalImage {
+            reader: self.reader.as_ref().unwrap().clone(),
+            start_offset: self.before_ablation_image_start_offset,
+            end_offset: self.before_ablation_image_end_offset,
+            image_format: ImageFormat::Png,
         }
+        // match self.dynamic_image(
+        //     self.before_ablation_image_start_offset,
+        //     self.before_ablation_image_end_offset,
+        // ) {
+        //     DynamicImage::ImageRgba8(rgba8) => rgba8,
+        //     _ => panic!("Unexpected DynamicImage type"),
+        // }
     }
 
     /// Returns the optical image of the acquisition region after ablation
-    pub fn after_ablation_image(&self) -> RgbaImage {
-        match self.dynamic_image(
-            self.after_ablation_image_start_offset,
-            self.after_ablation_image_end_offset,
-        ) {
-            DynamicImage::ImageRgba8(rgba8) => rgba8,
-            _ => panic!("Unexpected DynamicImage type"),
+    pub fn after_ablation_image(&self) -> OpticalImage<T> {
+        OpticalImage {
+            reader: self.reader.as_ref().unwrap().clone(),
+            start_offset: self.after_ablation_image_start_offset,
+            end_offset: self.after_ablation_image_end_offset,
+            image_format: ImageFormat::Png,
         }
     }
 
@@ -234,6 +289,71 @@ impl<T: Read + Seek> Acquisition<T> {
         expected_size == measured_size
     }
 
+    /// Returns a `Region` describing the pixel region contained within the specified bounding box
+    pub fn pixels_in(&self, region: &BoundingBox<f64>) -> Option<Region> {
+        let transform = self.to_slide_transform();
+
+        let top_left = transform.transform_from_slide(region.min_x, region.min_y)?;
+        let top_right = transform.transform_from_slide(region.max_x(), region.min_y)?;
+
+        let bottom_right = transform.transform_from_slide(region.max_x(), region.max_y())?;
+        let bottom_left = transform.transform_from_slide(region.min_x, region.max_y())?;
+
+        let min_x = top_left
+            .x
+            .min(top_right.x)
+            .min(bottom_right.x)
+            .min(bottom_left.x)
+            .max(0.0)
+            .floor();
+
+        let max_x = top_left
+            .x
+            .max(top_right.x)
+            .max(bottom_right.x)
+            .max(bottom_left.x)
+            .min(self.width() as f64)
+            .ceil();
+
+        let min_y = top_left
+            .y
+            .min(top_right.y)
+            .min(bottom_right.y)
+            .min(bottom_left.y)
+            .max(0.0)
+            .floor();
+
+        let max_y = top_left
+            .y
+            .max(top_right.y)
+            .max(bottom_right.y)
+            .max(bottom_left.y)
+            .min(self.height() as f64)
+            .ceil();
+
+        Some(Region {
+            x: min_x as u32,
+            y: (self.height() as u32 - max_y as u32),
+            width: (max_x - min_x) as u32,
+            height: (max_y - min_y) as u32,
+        })
+    }
+
+    /// Tests whether the acquisition is (at least partially) contained within the specified bounding box (slide coordinates).
+    pub fn in_region(&self, region: &BoundingBox<f64>) -> bool {
+        let slide_box = self.slide_bounding_box();
+
+        if slide_box.min_x < region.max_x()
+            && slide_box.max_x() > region.min_x
+            && slide_box.max_y() > region.min_y
+            && slide_box.min_y < region.max_y()
+        {
+            return true;
+        }
+
+        false
+    }
+
     /// Provides an iterator over all spectra (each pixel) within the acquisition
     pub fn spectra(&self) -> SpectrumIterator<T> {
         SpectrumIterator::new(self)
@@ -285,104 +405,158 @@ impl<T: Read + Seek> Acquisition<T> {
 
     /// Returns the ChannelImage for the channel matching the `ChannelIdentifier`. This contains the intensities of the channel
     /// for each detected pixel, the number of valid pixels and the width and height of the image.
-    pub fn channel_data(&self, identifier: &ChannelIdentifier) -> Result<ChannelImage, MCDError> {
-        let channel = self.channel(identifier).unwrap();
-        let mut data = vec![0.0f32; self.num_spectra()];
-        let mut min_value = f32::MAX;
-        let mut max_value = f32::MIN;
+    pub fn channel_image(
+        &self,
+        identifier: &ChannelIdentifier,
+        region: Option<Region>,
+    ) -> Result<ChannelImage, MCDError> {
+        self.channel_images(&[identifier.clone()], region)
+            .map(|mut data| data.drain(..).last().unwrap())
+    }
+
+    /// Returns array of ChannelImages for the channels matching the `ChannelIdentifier`s. This contains the intensities of the channel
+    /// for each detected pixel, the number of valid pixels and the width and height of the image.
+    pub fn channel_images(
+        &self,
+        identifiers: &[ChannelIdentifier],
+        region: Option<Region>,
+    ) -> Result<Vec<ChannelImage>, MCDError> {
+        // println!("Searching identifiers: {:?}", identifiers);
+        // println!("Searching from channels: {:?}", self.channels());
+
+        let channels: Option<Vec<_>> = identifiers
+            .iter()
+            .map(|identifier| self.channel(identifier))
+            .collect();
+
+        let channels = channels.ok_or(MCDError::NoSuchChannel {
+            acquisition: AcquisitionIdentifier::Id(self.id),
+        })?;
+
+        let order_numbers: Vec<_> = channels
+            .iter()
+            .map(|channel| channel.order_number() as usize)
+            .collect();
+
+        let region = match region {
+            Some(region) => region,
+            None => crate::Region {
+                x: 0,
+                y: 0,
+                width: self.width() as u32,
+                height: self.height() as u32,
+            },
+        };
+
+        let last_row = self.num_spectra() / self.width() as usize;
+        let last_col = self.width() as usize - (self.num_spectra() % self.width() as usize);
+
+        let valid_region_row = (region.y + region.height).min(last_row as u32);
+        let valid_region_col = (region.x + region.width).min(last_col as u32);
+
+        let valid_pixels =
+            ((valid_region_row - region.y - 1) * region.width) + (valid_region_col - region.x);
+
+        // println!(
+        //     "Valid pixels: {} | {} {} | {} {} | {}",
+        //     valid_pixels,
+        //     valid_region_col,
+        //     valid_region_row,
+        //     self.width(),
+        //     self.height(),
+        //     self.num_spectra()
+        // );
+
+        // let mut data = vec![0.0; valid_pixels as usize];
+        // let mut min_value = f32::MAX;
+        // let mut max_value = f32::MIN;
 
         //println!("DCM Location: {:?}", self.dcm_location);
 
         if let Some(data_location) = &self.dcm_location {
-            let mut reader = data_location.reader.lock().unwrap();
+            let mut data = data_location.read_channels(&order_numbers, &region)?;
 
-            let offset = data_location.offsets[channel.order_number() as usize];
-            let mut buf = vec![0; data_location.sizes[channel.order_number() as usize] as usize];
+            let images: Vec<_> = data
+                .drain(..)
+                //.zip(channels.iter())
+                .map(|data| {
+                    let mut min_value = f32::MAX;
+                    let mut max_value = f32::MIN;
 
-            /*println!("About to read channel {:?}", channel);
-            println!(
-                "[Order number: {}] Offset {} and buffer length {} with num spectra {}",
-                channel.order_number(),
-                offset,
-                buf.len(),
-                self.num_spectra()
-            );*/
+                    for &data_point in data.iter() {
+                        if data_point < min_value {
+                            min_value = data_point;
+                        }
+                        if data_point > max_value {
+                            max_value = data_point;
+                        }
+                    }
 
-            reader.seek(SeekFrom::Start(offset)).unwrap();
-            reader.read_exact(&mut buf).unwrap();
+                    ChannelImage {
+                        region,
+                        range: (min_value, max_value),
+                        valid_pixels: valid_pixels as usize,
+                        data,
+                    }
+                })
+                .collect();
 
-            let decompressed_data = lz4_flex::decompress(&buf, self.num_spectra() * 4);
-            if let Err(error) = decompressed_data {
-                return Err(MCDError::from(error));
-            }
-            let decompressed_data = decompressed_data.unwrap();
-            let mut decompressed_data = Cursor::new(decompressed_data);
-
-            decompressed_data
-                .read_f32_into::<LittleEndian>(&mut data)
-                .unwrap();
+            Ok(images)
         } else {
-            let offset = self.data_start_offset as u64
-                + channel.order_number() as u64 * self.value_bytes as u64;
+            todo!();
+            // let offset = self.data_start_offset as u64
+            //     + channel.order_number() as u64 * self.value_bytes as u64;
 
-            let mut reader = self.reader.as_ref().unwrap().lock().unwrap();
-            // TODO: Currently this only works for f32
-            let mut buf = [0; 4];
+            // let mut reader = self.reader.as_ref().unwrap().lock().unwrap();
+            // // TODO: Currently this only works for f32
+            // let mut buf = [0; 4];
 
-            // TODO: Handle this properly without unwrapping
-            reader.seek(SeekFrom::Start(offset)).unwrap();
-            for data_point in data.iter_mut() {
-                reader.read_exact(&mut buf).unwrap();
+            // reader.seek(SeekFrom::Start(offset))?;
+            // for data_point in data.iter_mut() {
+            //     reader.read_exact(&mut buf)?;
 
-                *data_point = f32::from_le_bytes(buf);
+            //     *data_point = f32::from_le_bytes(buf);
 
-                if *data_point > min_value {
-                    min_value = *data_point;
-                }
-                if *data_point < max_value {
-                    max_value = *data_point;
-                }
+            //     if *data_point < min_value {
+            //         min_value = *data_point;
+            //     }
+            //     if *data_point > max_value {
+            //         max_value = *data_point;
+            //     }
 
-                reader
-                    .seek(SeekFrom::Current(self.spectrum_size() as i64 - 4))
-                    .unwrap();
-            }
+            //     reader.seek(SeekFrom::Current(self.spectrum_size() as i64 - 4))?;
+            // }
         }
 
-        Ok(ChannelImage {
-            width: self.max_x,
-            height: self.max_y,
-            range: (min_value, max_value),
-            valid_pixels: self.num_spectra(),
-            data,
-        })
+        // Ok(ChannelImage {
+        //     region,
+        //     range: (min_value, max_value),
+        //     valid_pixels: valid_pixels as usize,
+        //     data,
+        // })
     }
 
     /// Returns the channel which matches the given identifier, or None if no match found
     pub fn channel(&self, identifier: &ChannelIdentifier) -> Option<&AcquisitionChannel> {
         for channel in &self.channels {
-            match identifier {
-                ChannelIdentifier::Order(order) => {
-                    if channel.order_number() == *order {
-                        return Some(channel);
-                    }
-                }
-                ChannelIdentifier::Name(name) => {
-                    if channel.name() == name {
-                        return Some(channel);
-                    }
-                }
-                ChannelIdentifier::Label(label) => {
-                    if channel.label() == label {
-                        return Some(channel);
-                    }
-                }
+            if channel.is(identifier) {
+                return Some(channel);
             }
         }
 
         None
     }
 
+    // pub fn channel_index(&self, identifier: &ChannelIdentifier) -> Option<usize> {
+    //     for (index, channel) in self.channels.iter().enumerate() {
+    //         if channel.is(identifier) {
+    //             return Some(index);
+    //         }
+    //     }
+
+    //     None
+    // }
+
     pub(crate) fn fix_roi_start_pos(&mut self) {
         // In version 2 of the schema, it seems like ROIStartXPosUm and ROIStartYPosUm are 1000x what they should be, so try and detect this and correct for it
         if self.roi_start_x_pos_um > 75000.0 {
@@ -396,7 +570,7 @@ impl<T: Read + Seek> Acquisition<T> {
     }
 }
 
-impl<T: Seek + Read> OnSlide for Acquisition<T> {
+impl<T: Seek + BufRead> OnSlide for Acquisition<T> {
     /// Returns the affine transformation from pixel coordinates within the acquisition to to the slide coordinates (μm)
     fn to_slide_transform(&self) -> AffineTransform<f64> {
         let mut moving_points = Vec::new();
@@ -413,26 +587,28 @@ impl<T: Seek + Read> OnSlide for Acquisition<T> {
 
         moving_points.push(Vector2::new(
             self.roi_start_x_pos_um,
-            25000.0 - self.roi_start_y_pos_um,
+            self.roi_start_y_pos_um,
+            //25000.0 - self.roi_start_y_pos_um,
         ));
         moving_points.push(Vector2::new(
             roi_end_x_pos_um,
-            25000.0 - self.roi_start_y_pos_um,
+            self.roi_start_y_pos_um,
+            //25000.0 - self.roi_start_y_pos_um,
         ));
         moving_points.push(Vector2::new(
             self.roi_start_x_pos_um,
-            25000.0 - self.roi_end_y_pos_um,
+            self.roi_end_y_pos_um,
+            //25000.0 - self.roi_end_y_pos_um,
         ));
         //moving_points.push(Vector2::new(roi_end_x_pos_um, self.roi_end_y_pos_um));
 
-        fixed_points.push(Vector2::new(0.0, 0.0));
-        fixed_points.push(Vector2::new(self.max_x as f64, 0.0));
+        let y = (self.roi_start_y_pos_um - self.roi_end_y_pos_um)
+            / self.ablation_distance_between_shots_y;
+
+        fixed_points.push(Vector2::new(0.0, y));
+        fixed_points.push(Vector2::new(self.max_x as f64, y));
         //fixed_points.push(Vector2::new(0.0, self.max_y as f64));
-        fixed_points.push(Vector2::new(
-            0.0,
-            (self.roi_start_y_pos_um - self.roi_end_y_pos_um)
-                / self.ablation_distance_between_shots_y,
-        ));
+        fixed_points.push(Vector2::new(0.0, 0.0));
 
         //fixed_points.push(Vector2::new(self.max_x as f64, self.max_y as f64));
 
@@ -460,7 +636,7 @@ impl<T: Seek + Read> OnSlide for Acquisition<T> {
 }
 
 #[rustfmt::skip]
-impl<T: Seek + Read> Print for Acquisition<T> {
+impl<T: Seek + BufRead> Print for Acquisition<T> {
     fn print<W: fmt::Write + ?Sized>(&self, writer: &mut W, indent: usize) -> fmt::Result {
         write!(writer, "{:indent$}", "", indent = indent)?;
         writeln!(writer, "{:-^1$}", "Acquisition", 48)?;
@@ -597,13 +773,13 @@ impl<T: Seek + Read> Print for Acquisition<T> {
     }
 }
 
-impl<T: Seek + Read> fmt::Display for Acquisition<T> {
+impl<T: Seek + BufRead> fmt::Display for Acquisition<T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         self.print(f, 0)
     }
 }
 
-impl<T: Seek + Read> From<AcquisitionXML> for Acquisition<T> {
+impl<T: Seek + BufRead> From<AcquisitionXML> for Acquisition<T> {
     fn from(acquisition: AcquisitionXML) -> Self {
         Acquisition {
             reader: None,
