@@ -22,39 +22,42 @@
 //! ```
 
 mod convert;
-mod error;
+/// Errors associated with parsing IMC data
+pub mod error;
 pub(crate) mod mcd;
-mod transform;
+/// Transformations (e.g. affine) used for converting
+pub mod transform;
 
 mod acquisition;
 mod calibration;
 mod channel;
-mod images;
 mod panorama;
 mod slide;
 
 /// Provides methods for reading in cell segmentation data from Halo
 pub mod halo;
 
-pub use self::acquisition::Acquisition;
+pub use self::acquisition::{Acquisition, AcquisitionIdentifier, Acquisitions};
 pub use self::channel::{AcquisitionChannel, ChannelIdentifier};
 pub use self::panorama::Panorama;
 pub use self::slide::Slide;
 
+use error::MCDError;
+use image::io::Reader as ImageReader;
 use std::convert::TryInto;
 use std::fmt;
-use std::io::{Read, Seek, SeekFrom, Write};
+use std::io::{BufRead, Cursor, Seek, SeekFrom};
 
+use std::ops::DerefMut;
 use std::path::PathBuf;
 use std::sync::{Arc, Mutex};
 
 use std::collections::HashMap;
 
-use acquisition::AcquisitionIdentifier;
 use calibration::{Calibration, CalibrationChannel, CalibrationFinal, CalibrationParams};
 use mcd::{MCDParser, ParserState};
 
-use image::{ImageFormat, RgbaImage};
+use image::{DynamicImage, ImageFormat, ImageResult, RgbaImage};
 use slide::{SlideFiducialMarks, SlideProfile};
 use transform::AffineTransform;
 
@@ -67,15 +70,79 @@ pub trait Print {
 }
 
 /// Represents property of having an optical image
-pub trait OpticalImage {
+pub struct OpticalImage<T: Seek + BufRead> {
+    reader: Arc<Mutex<T>>,
+
+    start_offset: i64,
+    end_offset: i64,
+    image_format: ImageFormat,
+}
+
+impl<T: Seek + BufRead> OpticalImage<T> {
     /// Returns whether an optical image is present
-    fn has_image(&self) -> bool;
+    //fn has_image(&self) -> bool;
     /// Returns the binary data for the image, exactly as stored in the .mcd file
-    fn image_data(&self) -> Result<Vec<u8>, std::io::Error>;
+    pub fn image_data(&self) -> Result<Vec<u8>, std::io::Error> {
+        let mut reader = self.reader.lock().unwrap();
+
+        let start_offset = self.start_offset();
+        let image_size = self.image_size().try_into().unwrap();
+
+        let mut buf_u8 = vec![0; image_size];
+
+        match reader.seek(SeekFrom::Start(start_offset as u64)) {
+            Ok(_seek) => match reader.read_exact(&mut buf_u8) {
+                Ok(()) => Ok(buf_u8),
+                Err(error) => Err(error),
+            },
+            Err(error) => Err(error),
+        }
+    }
+
+    fn start_offset(&self) -> i64 {
+        self.start_offset + 161
+    }
+
+    fn image_size(&self) -> i64 {
+        self.end_offset - self.start_offset()
+    }
+
     /// Returns the format of the stored optical image
-    fn image_format(&self) -> ImageFormat;
+    pub fn image_format(&self) -> ImageFormat {
+        self.image_format
+    }
+
     /// Returns a decoded RgbaImage of the panorama image
-    fn image(&self) -> RgbaImage;
+    pub fn as_rgba8(&self) -> Result<RgbaImage, std::io::Error> {
+        // match self.dynamic_image()? {
+        //     DynamicImage::ImageRgba8(rgba8) => Ok(rgba8),
+        //     DynamicImage::ImageRgb8(rgb8) => Ok(DynamicImage::ImageRgb8(rgb8).into_rgba8()),
+        //     _ => panic!("Unexpected DynamicImage type"),
+        // }
+        Ok(self.dynamic_image()?.into_rgba8())
+    }
+
+    /// Returns the dimensions of the images in pixels as a tuple (width, height)
+    pub fn dimensions(&self) -> ImageResult<(u32, u32)> {
+        let mut guard = self.reader.lock().unwrap();
+        let reader: &mut T = guard.deref_mut();
+
+        let start_offset = self.start_offset();
+        reader.seek(SeekFrom::Start(start_offset as u64))?;
+
+        let mut reader = ImageReader::new(reader);
+        reader.set_format(self.image_format());
+
+        reader.into_dimensions()
+    }
+
+    fn dynamic_image(&self) -> Result<DynamicImage, std::io::Error> {
+        let mut reader = ImageReader::new(Cursor::new(self.image_data()?));
+        reader.set_format(self.image_format);
+
+        // TODO: Deal with this error properly
+        Ok(reader.decode().unwrap())
+    }
 }
 
 /// Represents an image which is acquired on a slide
@@ -86,9 +153,22 @@ pub trait OnSlide {
     fn to_slide_transform(&self) -> AffineTransform<f64>;
 }
 
+/// Represents a region of an image (in pixels)
+#[derive(Debug, Clone, Copy)]
+pub struct Region {
+    /// x-position of the top left corner of the region
+    pub x: u32,
+    /// y-position of the top left corner of the region
+    pub y: u32,
+    /// width of the region in pixels
+    pub width: u32,
+    /// height of the region in pixels
+    pub height: u32,
+}
+
 /// Represents a imaging mass cytometry (*.mcd) file.
 #[derive(Debug)]
-pub struct MCD<T: Seek + Read> {
+pub struct MCD<T: Seek + BufRead> {
     reader: Arc<Mutex<T>>,
     location: String,
 
@@ -123,7 +203,7 @@ fn u16_from_u8(a: &mut [u16], v: &[u8]) {
     }
 }
 
-impl<T: Seek + Read> MCD<T> {
+impl<T: Seek + BufRead> MCD<T> {
     fn new(reader: T, location: &str) -> Self {
         MCD {
             reader: Arc::new(Mutex::new(reader)),
@@ -144,6 +224,11 @@ impl<T: Seek + Read> MCD<T> {
         }
     }
 
+    /// Returns the location (path) of the .mcd file
+    pub fn location(&self) -> &str {
+        &self.location
+    }
+
     pub(crate) fn dcm_file(&self) -> PathBuf {
         let mut path = PathBuf::from(&self.location);
         path.set_extension("dcm");
@@ -245,6 +330,23 @@ impl<T: Seek + Read> MCD<T> {
         None
     }
 
+    /// Returns a list of acquisitions which are at least partially contained within the specified bounding box.
+    pub fn acquisitions_in(&self, region: &BoundingBox<f64>) -> Vec<&Acquisition<T>> {
+        let mut acquisitions = Vec::new();
+
+        for slide in self.slides.values() {
+            for panorama in slide.panoramas() {
+                for acquisition in panorama.acquisitions() {
+                    if acquisition.in_region(region) {
+                        acquisitions.push(acquisition);
+                    }
+                }
+            }
+        }
+
+        acquisitions
+    }
+
     /// Returns a vector of all channels present within any acquisition performed on the slide, sorted by channel order number.
     pub fn channels(&self) -> Vec<&AcquisitionChannel> {
         let mut channels = HashMap::new();
@@ -267,7 +369,40 @@ impl<T: Seek + Read> MCD<T> {
             ordered_channels.push(channel);
         }
 
-        ordered_channels.sort_by_key(|a| a.order_number());
+        ordered_channels.sort_by_key(|a| a.label());
+
+        ordered_channels
+    }
+
+    /// Returns a vector of all channels, excluding those from the acquisitions with names matching those specified
+    pub fn channels_excluding(&self, exclusion_list: Vec<&str>) -> Vec<&AcquisitionChannel> {
+        let mut channels = HashMap::new();
+
+        // This should be unnecessary - hopefully there is only one set of channels per dataset?
+        for slide in self.slides.values() {
+            for panorama in slide.panoramas() {
+                for acquisition in panorama.acquisitions() {
+                    if !exclusion_list.contains(&acquisition.description()) {
+                        for channel in acquisition.channels() {
+                            if !channels.contains_key(channel.name())
+                                && channel.name() != "X"
+                                && channel.name() != "Y"
+                                && channel.name() != "Z"
+                            {
+                                channels.insert(channel.name(), channel);
+                            }
+                        }
+                    }
+                }
+            }
+        }
+
+        let mut ordered_channels = Vec::new();
+        for (_, channel) in channels.drain() {
+            ordered_channels.push(channel);
+        }
+
+        ordered_channels.sort_by_key(|a| a.label().to_ascii_lowercase());
 
         ordered_channels
     }
@@ -303,12 +438,12 @@ impl<T: Seek + Read> MCD<T> {
     }
 
     /// Parse *.mcd format
-    pub fn parse(reader: T, location: &str) -> Self {
-        let mut mcd = MCD::new(reader, location);
-        let combined_xml = mcd.xml();
+    pub fn parse(reader: T, location: &str) -> std::io::Result<Self> {
+        let mcd = MCD::new(reader, location);
+        let combined_xml = mcd.xml()?;
 
-        let mut file = std::fs::File::create("tmp.xml").unwrap();
-        file.write_all(combined_xml.as_bytes()).unwrap();
+        // let mut file = std::fs::File::create("tmp.xml").unwrap();
+        // file.write_all(combined_xml.as_bytes())?;
 
         let mut parser = MCDParser::new(mcd);
 
@@ -347,7 +482,7 @@ impl<T: Seek + Read> MCD<T> {
             buf.clear();
         }
 
-        parser.mcd()
+        Ok(parser.mcd())
     }
 
     /// Parse *.mcd format where a temporary file is generated for faster access to channel images.
@@ -355,20 +490,20 @@ impl<T: Seek + Read> MCD<T> {
     /// Data is stored in the *.mcd file spectrum-wise which means to load a single image, the entire acquired acquisition must be loaded first.
     /// This method creates a temporary file (*.dcm) in the same location as the *.mcd file (if it doesn't already exist) which has the channel
     /// data stored image-wise. If this file is present and loaded, then `Mcd` will choose the fastest method to use to return the requested data.
-    pub fn parse_with_dcm(reader: T, location: &str) -> Self {
-        let mut mcd = Self::parse(reader, location);
+    pub fn parse_with_dcm(reader: T, location: &str) -> Result<Self, MCDError> {
+        let mut mcd = Self::parse(reader, location)?;
 
         if std::fs::metadata(mcd.dcm_file()).is_err() {
             convert::convert(&mcd).unwrap();
         }
 
-        convert::open(&mut mcd).unwrap();
+        convert::open(&mut mcd)?;
 
-        mcd
+        Ok(mcd)
     }
 
     /// Returns the raw XML metadata stored in the .mcd file
-    pub fn xml(&mut self) -> String {
+    pub fn xml(&self) -> std::io::Result<String> {
         let chunk_size: i64 = 1000;
         let mut cur_offset: i64 = 0;
 
@@ -377,11 +512,9 @@ impl<T: Seek + Read> MCD<T> {
         loop {
             let mut reader = self.reader.lock().unwrap();
 
-            reader
-                .seek(SeekFrom::End(-cur_offset - chunk_size))
-                .unwrap();
+            reader.seek(SeekFrom::End(-cur_offset - chunk_size))?;
 
-            reader.read_exact(&mut buf_u8).unwrap();
+            reader.read_exact(&mut buf_u8)?;
 
             match std::str::from_utf8(&buf_u8) {
                 Ok(_data) => {}
@@ -392,8 +525,8 @@ impl<T: Seek + Read> MCD<T> {
                     let total_size = cur_offset + chunk_size - (start_index as i64);
                     buf_u8 = vec![0; total_size.try_into().unwrap()];
 
-                    reader.seek(SeekFrom::End(-total_size)).unwrap();
-                    reader.read_exact(&mut buf_u8).unwrap();
+                    reader.seek(SeekFrom::End(-total_size))?;
+                    reader.read_exact(&mut buf_u8)?;
 
                     break;
                 }
@@ -414,12 +547,12 @@ impl<T: Seek + Read> MCD<T> {
             }
         }
 
-        combined_xml
+        Ok(combined_xml)
     }
 }
 
 #[rustfmt::skip]
-impl<T: Seek + Read> Print for MCD<T> {
+impl<T: Seek + BufRead> Print for MCD<T> {
     fn print<W: fmt::Write + ?Sized>(&self, writer: &mut W, indent: usize) -> fmt::Result {
         writeln!(writer, "MCD File: {}", self.location)?;
 
@@ -452,7 +585,7 @@ impl<T: Seek + Read> Print for MCD<T> {
     }
 }
 
-impl<T: Seek + Read> fmt::Display for MCD<T> {
+impl<T: Seek + BufRead> fmt::Display for MCD<T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         self.print(f, 0)
     }
@@ -466,8 +599,8 @@ pub enum ImageFormat {
 }*/
 
 /// Represents a bounding rectangle
-#[derive(Debug)]
-pub struct BoundingBox<T> {
+#[derive(Debug, Clone)]
+pub struct BoundingBox<T: num_traits::Num + Copy> {
     /// Minimum x coordinate for the bounding rectangle
     pub min_x: T,
     /// Minimum y coordinate for the bounding rectangle
@@ -478,11 +611,23 @@ pub struct BoundingBox<T> {
     pub height: T,
 }
 
+impl<T: num_traits::Num + Copy> BoundingBox<T> {
+    /// Maximum x coordinate for the bounding rectangle
+    pub fn max_x(&self) -> T {
+        self.min_x + self.width
+    }
+
+    /// Maximum y coordinate for the bounding rectangle
+    pub fn max_y(&self) -> T {
+        self.min_y + self.height
+    }
+}
+
 /// Represents a channel image (stored as a vector of f32).
 /// If the run was stopped mid acquisition width*height != valid_pixels
 pub struct ChannelImage {
-    width: i32,
-    height: i32,
+    region: Region,
+
     range: (f32, f32),
     valid_pixels: usize,
     data: Vec<f32>,
@@ -490,13 +635,13 @@ pub struct ChannelImage {
 
 impl ChannelImage {
     /// Returns the width (in pixels) of the image
-    pub fn width(&self) -> i32 {
-        self.width
+    pub fn width(&self) -> u32 {
+        self.region.width
     }
 
     /// Returns the height (in pixels) of the image
-    pub fn height(&self) -> i32 {
-        self.height
+    pub fn height(&self) -> u32 {
+        self.region.height
     }
 
     /// Returns a pair (min, max) of f32 describing the limits of the detected intensities in the image
@@ -506,7 +651,7 @@ impl ChannelImage {
 
     /// Returns whether the data is complete (true) or whether the data acquisition aborted (false)
     pub fn is_complete(&self) -> bool {
-        self.valid_pixels == (self.width * self.height) as usize
+        self.valid_pixels == (self.region.width * self.region.height) as usize
     }
 
     /// Returns the number of valid pixels in the image. If the run was aborted part way through `num_valid_pixels() < width() * height()`
@@ -522,15 +667,18 @@ impl ChannelImage {
 
 #[cfg(test)]
 mod tests {
+    use image::{ImageBuffer, Pixel, Rgba};
+
     use super::*;
 
     use std::fs::File;
 
     use std::io::BufReader;
+    use std::time::Instant;
 
     #[test]
-    fn test_all_in_folder() -> std::io::Result<()> {
-        let paths = std::fs::read_dir("../test/").unwrap();
+    fn test_all_in_folder() -> Result<(), MCDError> {
+        let paths = std::fs::read_dir("/home/alan/Documents/Work/Nicole/Salmonella").unwrap();
 
         for path in paths {
             let path = path?;
@@ -541,9 +689,13 @@ mod tests {
             }
 
             let file = BufReader::new(File::open(path.path()).unwrap());
-            let mut mcd = MCD::parse_with_dcm(file, path.path().to_str().unwrap());
+            let mcd = MCD::parse_with_dcm(file, path.path().to_str().unwrap())?;
 
-            let _xml = mcd.xml();
+            // let overview_image = mcd.slides()[0].create_overview_image(7500, None).unwrap();
+
+            // overview_image.save("overview.png").unwrap();
+
+            //let _xml = mcd.xml()?;
 
             //println!("{}", xml);
 
@@ -560,12 +712,12 @@ mod tests {
                 .unwrap();
 
             let x_channel = acquisition
-                .channel_data(&ChannelIdentifier::Name("X".to_string()))
+                .channel_image(&ChannelIdentifier::Name("X".to_string()), None)
                 .unwrap();
 
             println!("Loaded X Channel : {:?}", x_channel.num_valid_pixels());
 
-            for channel in acquisition.channels() {
+            for channel in mcd.channels_excluding(vec!["ROI 12", "ROI 13", "ROI 14", "ROI 16"]) {
                 println!(
                     "[Channel {}] {} | {}",
                     channel.id(),
@@ -575,13 +727,73 @@ mod tests {
             }
 
             let channel_identifier = ChannelIdentifier::Name("Ir(191)".to_string());
-            let overview_image = mcd.slides()[0]
-                .create_overview_image(15000, Some((&channel_identifier, Some(100.0))))
-                .unwrap();
+            println!("Subimage");
+            let data = acquisition.channel_images(
+                &[channel_identifier.clone()],
+                // None,
+                Some(Region {
+                    x: 1000,
+                    y: 1000,
+                    width: 500,
+                    height: 500,
+                }),
+            )?;
 
-            overview_image.save("overview.png").unwrap();
+            let data = &data[0];
+
+            let mut acq_image: ImageBuffer<Rgba<u8>, Vec<u8>> =
+                ImageBuffer::new(data.width(), data.height());
+
+            let mut index = 0;
+            let max_value = 20.0;
+            for y in 0..data.height() {
+                if index >= data.valid_pixels {
+                    break;
+                }
+
+                for x in 0..data.width() {
+                    if index >= data.valid_pixels {
+                        break;
+                    }
+
+                    let g = ((data.data[index] / max_value) * 255.0) as u8;
+                    let g = g as f64 / 255.0;
+
+                    let cur_pixel = acq_image.get_pixel_mut(x as u32, y as u32).channels_mut();
+                    cur_pixel[1] = (g * 255.0) as u8;
+                    cur_pixel[3] = 255;
+
+                    index += 1;
+                }
+            }
+
+            acq_image.save("dna.png").unwrap();
+
+            let start = Instant::now();
+            let mut image_map = HashMap::new();
+
+            for acquisition in mcd.acquisitions() {
+                if let Ok(data) = acquisition.channel_image(
+                    &channel_identifier,
+                    // None
+                    Some(Region {
+                        x: 1000,
+                        y: 1000,
+                        width: 500,
+                        height: 500,
+                    }),
+                ) {
+                    image_map.insert(format!("{}", acquisition.id()), ChannelImage(data));
+                }
+            }
+
+            let duration = start.elapsed();
+
+            println!("Time elapsed loading data is: {:?}", duration);
         }
 
         Ok(())
     }
+
+    pub struct ChannelImage(crate::ChannelImage);
 }
