@@ -1,19 +1,16 @@
 use core::fmt;
 use std::{
     collections::HashMap,
-    io::{Cursor, Read, Seek},
+    io::{BufRead, Seek},
     sync::{Arc, Mutex},
 };
 
-use image::{
-    imageops::FilterType, DynamicImage, ImageBuffer, ImageFormat, RgbImage, Rgba, RgbaImage,
-};
-use image::{io::Reader as ImageReader, Pixel};
+use image::Pixel;
+use image::{imageops::FilterType, ImageBuffer, ImageFormat, Rgba, RgbaImage};
 
 use crate::{
     channel::ChannelIdentifier,
     error::MCDError,
-    images::read_image_data,
     mcd::{SlideFiducialMarksXML, SlideProfileXML},
     OnSlide, OpticalImage, Panorama, Print,
 };
@@ -22,7 +19,7 @@ use crate::mcd::SlideXML;
 
 /// Represents a slide (contains multiple panoramas) in the *.mcd format
 #[derive(Debug)]
-pub struct Slide<T: Seek + Read> {
+pub struct Slide<T: Seek + BufRead> {
     pub(crate) reader: Option<Arc<Mutex<T>>>,
 
     id: u16,
@@ -50,7 +47,7 @@ pub struct Slide<T: Seek + Read> {
     panoramas: HashMap<u16, Panorama<T>>,
 }
 
-impl<T: Seek + Read> From<SlideXML> for Slide<T> {
+impl<T: Seek + BufRead> From<SlideXML> for Slide<T> {
     fn from(slide: SlideXML) -> Self {
         Slide {
             reader: None,
@@ -78,7 +75,7 @@ impl<T: Seek + Read> From<SlideXML> for Slide<T> {
     }
 }
 
-impl<T: Seek + Read> Slide<T> {
+impl<T: Seek + BufRead> Slide<T> {
     /// Returns the slide ID
     pub fn id(&self) -> u16 {
         self.id
@@ -148,34 +145,49 @@ impl<T: Seek + Read> Slide<T> {
     }
 
     /// Returns associated image data
-    pub fn image_data(&self) -> Result<Vec<u8>, std::io::Error> {
-        let mutex = self
-            .reader
-            .as_ref()
-            .expect("Should have copied the reader across");
-        let reader = mutex.lock().unwrap();
+    // pub fn image_data(&self) -> Result<Vec<u8>, std::io::Error> {
+    //     let mutex = self
+    //         .reader
+    //         .as_ref()
+    //         .expect("Should have copied the reader across");
+    //     let reader = mutex.lock().unwrap();
 
-        read_image_data(reader, self.image_start_offset, self.image_end_offset)
-    }
+    //     read_image_data(reader, self.image_start_offset, self.image_end_offset)
+    // }
 
-    fn dynamic_image(&self) -> DynamicImage {
-        let mut reader = ImageReader::new(Cursor::new(self.image_data().unwrap()));
+    /// Returns the format describing the binary image data
+    fn image_format(&self) -> ImageFormat {
         if self.software_version().starts_with('6') {
-            reader.set_format(ImageFormat::Jpeg);
+            ImageFormat::Jpeg
         } else {
-            reader.set_format(ImageFormat::Png);
+            ImageFormat::Png
         }
-        reader.decode().unwrap()
     }
 
-    /// Returns the image associated with the slide.
-    pub fn image(&self) -> RgbImage {
-        match self.dynamic_image() {
-            DynamicImage::ImageRgb8(rgb8) => rgb8,
-            _ => panic!("Unexpected DynamicImage type"),
+    /// Returns the image associated with the slide
+    pub fn image(&self) -> OpticalImage<T> {
+        OpticalImage {
+            reader: self.reader.as_ref().unwrap().clone(),
+            start_offset: self.image_start_offset,
+            end_offset: self.image_end_offset,
+            image_format: self.image_format(),
         }
     }
 
+    // fn dynamic_image(&self) -> DynamicImage {
+    //     let mut reader = ImageReader::new(Cursor::new(self.image_data().unwrap()));
+    //     reader.set_format(self.image_format());
+    //     reader.decode().unwrap()
+    // }
+
+    // /// Returns the image associated with the slide.
+    // pub fn image(&self) -> RgbImage {
+    //     match self.dynamic_image() {
+    //         DynamicImage::ImageRgb8(rgb8) => rgb8,
+    //         _ => panic!("Unexpected DynamicImage type"),
+    //     }
+    // }
+
     /// Create an overview image of the slide scaled to the supplied width.
     ///
     /// This will scale the slide image to the supplied width, and overlay any panorama images acquired.
@@ -186,7 +198,7 @@ impl<T: Seek + Read> Slide<T> {
         width: u32,
         channel_to_show: Option<(&ChannelIdentifier, Option<f32>)>,
     ) -> Result<RgbaImage, MCDError> {
-        let slide_image = self.dynamic_image();
+        let slide_image = self.image().dynamic_image().unwrap();
 
         // Move into function to help debugging
         /*match &slide_image {
@@ -208,16 +220,21 @@ impl<T: Seek + Read> Slide<T> {
 
         println!("Saved !");*/
 
+        let ratio = self.height_in_um() / self.width_in_um();
+        let output_image_height = (width as f64 * ratio) as u32;
+
         let mut resized_image = slide_image
             .resize_exact(width, width / 3, FilterType::Nearest)
             .to_rgba8();
         //println!("Resized !");
 
-        let scale = 75000.0 / width as f64;
+        //return Ok(slide_image.to_rgba8());
+
+        let scale = self.width_in_um() / width as f64;
 
         for panorama in self.panoramas() {
             if panorama.has_image() {
-                let panorama_image = panorama.image();
+                let panorama_image = panorama.image().unwrap().as_rgba8().unwrap();
 
                 //let panorama_image = panorama_image.to_rgba8();
 
@@ -254,7 +271,7 @@ impl<T: Seek + Read> Slide<T> {
                             .unwrap();
 
                         let pixel_x = new_point[0].round() as i32;
-                        let pixel_y = height as i32 - new_point[1].round() as i32;
+                        let pixel_y = panorama_image.height() as i32 - new_point[1].round() as i32;
 
                         if pixel_x < 0
                             || pixel_y < 0
@@ -266,7 +283,7 @@ impl<T: Seek + Read> Slide<T> {
 
                         let pixel = *panorama_image.get_pixel(pixel_x as u32, pixel_y as u32);
 
-                        resized_image.put_pixel(x, y, pixel);
+                        resized_image.put_pixel(x, output_image_height - y, pixel);
                     }
                 }
 
@@ -292,7 +309,7 @@ impl<T: Seek + Read> Slide<T> {
 
                     //let bounding_box = acquisition.slide_bounding_box();
                     let transform = acquisition.to_slide_transform();
-                    let data = acquisition.channel_data(identifier)?;
+                    let data = acquisition.channel_image(identifier, None)?;
 
                     let max_value = match max_value {
                         Some(value) => value,
@@ -302,14 +319,14 @@ impl<T: Seek + Read> Slide<T> {
                     let mut index = 0;
 
                     let mut acq_image: ImageBuffer<Rgba<u8>, Vec<u8>> =
-                        ImageBuffer::new(data.width as u32, data.height as u32);
+                        ImageBuffer::new(data.width(), data.height());
 
-                    for y in 0..data.height {
+                    for y in 0..data.height() {
                         if index >= data.valid_pixels {
                             break;
                         }
 
-                        for x in 0..data.width {
+                        for x in 0..data.width() {
                             if index >= data.valid_pixels {
                                 break;
                             }
@@ -397,7 +414,7 @@ impl<T: Seek + Read> Slide<T> {
 }
 
 #[rustfmt::skip]
-impl<T: Seek + Read> Print for Slide<T> {
+impl<T: Seek + BufRead> Print for Slide<T> {
     fn print<W: fmt::Write + ?Sized>(&self, writer: &mut W, indent: usize) -> fmt::Result {
         write!(writer, "{:indent$}", "", indent = indent)?;
         writeln!(writer, "{:-^1$}", "Slide", 36)?;
@@ -489,7 +506,7 @@ impl<T: Seek + Read> Print for Slide<T> {
     }
 }
 
-impl<T: Seek + Read> fmt::Display for Slide<T> {
+impl<T: Seek + BufRead> fmt::Display for Slide<T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         self.print(f, 0)
     }
